#include <stdio.h>
#include "opencv2/opencv.hpp"

#define PIXVAL(iplimagep, x, y) (*(uchar *)((iplimagep)->imageData + (y) * (iplimagep)->widthStep + (x)))
#define PIXVAL_16S(iplimagep, x, y) (*(signed short *)((iplimagep)->imageData + (y) * (iplimagep)->widthStep + (2 * (x))))

#define B_SIZE 64
#define EIG_THRESH (B_SIZE * B_SIZE * 3)
#define MAX_ITER 20

void paste_box(IplImage *img, IplImage *panel, int x, int y, int b_size, double scale)
{
    IplImage *tmp = cvCreateImage(cvGetSize(img), IPL_DEPTH_16S, 1);
    cvConvertScale(img, tmp, scale);
    cvSetImageROI(panel, cvRect(x, y, b_size, b_size));
    cvCopy(tmp, panel);
    cvResetImageROI(panel);
    cvRectangle(panel, cvPoint(x - 1, y - 1), cvPoint(x + b_size, y + b_size),
                cvScalar(32767), 1);
    cvReleaseImage(&tmp);
}

void put_text_center(IplImage *img, char *text, CvPoint center, CvFont *font, CvScalar color)
{
    CvSize sz;
    int baseline;

    cvGetTextSize(text, font, &sz, &baseline);
    cvPutText(img, text,
              cvPoint(center.x - sz.width / 2, center.y + sz.height / 2),
              font, color);
}

int main()
{
    int i, j;

    // block size and block boundaries
    int b_size = B_SIZE;
    int b_xmin, b_xmax, b_ymin, b_ymax;
    float b_xcen, b_ycen;
    
    IplImage *frame, *buf;
    IplImage *I, *J, *Iblock, *Jblock;
    IplImage *Ix, *Iy, *Ix2, *Iy2, *IxIy, *IminJ, *IminJ_Ix, *IminJ_Iy;
    IplImage *panel;
    CvMat *G, *e, *eig, *sol;
    CvFont font;

    CvCapture *capture = cvCaptureFromCAM(0);
    if (capture == NULL) {
        fprintf(stderr, "cannot capture\n");
        return 1;
    }

    frame = cvQueryFrame(capture);
    buf = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);  // color

    I = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
    J = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
    Iblock = cvCreateImage(cvSize(b_size, b_size), IPL_DEPTH_8U, 1);
    Jblock = cvCreateImage(cvSize(b_size, b_size), IPL_DEPTH_8U, 1);

    Ix = cvCreateImage(cvSize(b_size, b_size), IPL_DEPTH_16S, 1);
    Iy = cvCreateImage(cvSize(b_size, b_size), IPL_DEPTH_16S, 1);
    Ix2 = cvCreateImage(cvSize(b_size, b_size), IPL_DEPTH_16S, 1);
    Iy2 = cvCreateImage(cvSize(b_size, b_size), IPL_DEPTH_16S, 1);
    IxIy = cvCreateImage(cvSize(b_size, b_size), IPL_DEPTH_16S, 1);
    IminJ = cvCreateImage(cvSize(b_size, b_size), IPL_DEPTH_16S, 1);
    IminJ_Ix = cvCreateImage(cvSize(b_size, b_size), IPL_DEPTH_16S, 1);
    IminJ_Iy = cvCreateImage(cvSize(b_size, b_size), IPL_DEPTH_16S, 1);

    panel = cvCreateImage(cvSize(b_size * 8, b_size * 3.5),
                          IPL_DEPTH_16S, 1);

    cvNamedWindow("image", CV_WINDOW_AUTOSIZE);
    cvNamedWindow("panel", CV_WINDOW_AUTOSIZE);

    b_xmin = (I->width / 2) - (b_size / 2);
    b_xmax = b_xmin + b_size - 1;
    b_ymin = (I->height / 2) - (b_size / 2);
    b_ymax = b_ymin + b_size - 1;
    b_xcen = (b_xmin + b_xmax) / 2.0;
    b_ycen = (b_ymin + b_ymax) / 2.0;

    G = cvCreateMat(2, 2, CV_32F);
    e = cvCreateMat(2, 1, CV_32F);
    sol = cvCreateMat(2, 1, CV_32F);
    eig = cvCreateMat(2, 1, CV_32F);

    cvZero(I);

    while (1) {
        float dx = 0, dy = 0, dx_prev = 0, dy_prev = 0;
        float dx_traj[MAX_ITER], dy_traj[MAX_ITER];
        float G11 = 0, G12 = 0, G22 = 0;
        float eig1, eig2, min_eig;
        int is_converged = 0;
        int iter;
        
        frame = cvQueryFrame(capture); /* cannot be modified */
        cvCopy(frame, buf);
        cvCvtColor(buf, J, CV_BGR2GRAY);

        for (j = 0; j < b_size; j++) {
            for (i = 0; i <= b_size; i++) {
                int bi = b_xmin + i;
                int bj = b_ymin + j;

                PIXVAL(Iblock, i, j) = PIXVAL(I, bi, bj);

                PIXVAL_16S(Ix, i, j)
                    = (PIXVAL(I, bi + 1, bj) - PIXVAL(I, bi - 1, bj)) / 2;
                PIXVAL_16S(Iy, i, j)
                    = (PIXVAL(I, bi, bj + 1) - PIXVAL(I, bi, bj - 1)) / 2;

                PIXVAL_16S(Ix2, i, j)
                    = PIXVAL_16S(Ix, i, j) * PIXVAL_16S(Ix, i, j);
                G11 += PIXVAL_16S(Ix2, i, j);
                PIXVAL_16S(IxIy, i, j)
                    = PIXVAL_16S(Ix, i, j) * PIXVAL_16S(Iy, i, j);
                G12 += PIXVAL_16S(IxIy, i, j);
                PIXVAL_16S(Iy2, i, j)
                    = PIXVAL_16S(Iy, i, j) * PIXVAL_16S(Iy, i, j);
                G22 += PIXVAL_16S(Iy2, i, j);
            }
        }

        cvmSet(G, 0, 0, G11); cvmSet(G, 0, 1, G12);
        cvmSet(G, 1, 0, G12); cvmSet(G, 1, 1, G22);
        cvSVD(G, eig);
        eig1 = cvmGet(eig, 0, 0);
        eig2 = cvmGet(eig, 1, 0);
        min_eig = MIN(eig1, eig2);

        if (min_eig < EIG_THRESH) {
            // just for displaying
            cvZero(IminJ_Ix);
            cvZero(IminJ_Iy);
        } else {
            for (iter = 0; iter < MAX_ITER && !is_converged; iter++) {
                float e1 = 0, e2 = 0;
                cvGetRectSubPix(J, Jblock,
                                cvPoint2D32f(b_xcen + dx, b_ycen + dy));
                for (j = 0; j < b_size; j++) {
                    for (i = 0; i <= b_size; i++) {
                        PIXVAL_16S(IminJ, i, j)
                            = PIXVAL(Iblock, i, j) - PIXVAL(Jblock, i, j);
                        PIXVAL_16S(IminJ_Ix, i, j)
                            = PIXVAL_16S(IminJ, i, j) * PIXVAL_16S(Ix, i, j);
                        e1 += PIXVAL_16S(IminJ_Ix, i, j);
                        PIXVAL_16S(IminJ_Iy, i, j)
                            = PIXVAL_16S(IminJ, i, j) * PIXVAL_16S(Iy, i, j);
                        e2 += PIXVAL_16S(IminJ_Iy, i, j);
                    }
                }

                cvmSet(e, 0, 0, e1); 
                cvmSet(e, 1, 0, e2);
                cvSolve(G, e, sol);

                dx_prev = dx;
                dy_prev = dy;
                dx += cvmGet(sol, 0, 0);
                dy += cvmGet(sol, 1, 0);
                dx_traj[iter] = dx;
                dy_traj[iter] = dy;

                is_converged = ((dx - dx_prev)*(dx - dx_prev) + (dy - dy_prev)*(dy - dy_prev) < 0.05)? 1: 0;

            }
        }
        
        cvCopy(J, I);


        /*
         *  Differential Image Panel
         */

        cvSet(panel, cvScalar(-32768));

        int u = (b_size / 4 >= 2)? (b_size / 4): 2;

        // matrix brackets
        cvRectangle(panel, cvPoint(u, u), cvPoint(13 * u, 13 * u),
                    cvScalar(32767), 1);
        cvRectangle(panel, cvPoint(2 * u, u), cvPoint(12 * u, 13 * u),
                    cvScalar(-32768), 1);
        cvRectangle(panel, cvPoint(15 * u, u), cvPoint(21 * u, 13 * u),
                    cvScalar(32767), 1);
        cvRectangle(panel, cvPoint(16 * u, u), cvPoint(20 * u, 13 * u),
                    cvScalar(-32768), 1);
        cvRectangle(panel, cvPoint(25 * u, u), cvPoint(31 * u, 13 * u),
                    cvScalar(32767), 1);
        cvRectangle(panel, cvPoint(26 * u, u), cvPoint(30 * u, 13 * u),
                    cvScalar(-32768), 1);
        // texts
        cvInitFont(&font, CV_FONT_HERSHEY_PLAIN, 1.0, 1.0);
        put_text_center(panel, "dx", cvPoint(18 * u, 4 * u),
                        &font, cvScalar(32767));
        put_text_center(panel, "dy", cvPoint(18 * u, 10 * u),
                        &font, cvScalar(32767));
        put_text_center(panel, "=", cvPoint(23 * u, 7 * u),
                        &font, cvScalar(32767));

        // differential images
        paste_box(Ix2, panel, 2 * u, 2 * u, b_size, 50);
        paste_box(IxIy, panel, 8 * u, 2 * u, b_size, 50);
        paste_box(IxIy, panel, 2 * u, 8 * u, b_size, 50);
        paste_box(Iy2, panel, 8 * u, 8 * u, b_size, 50);
        paste_box(IminJ_Ix, panel, 26 * u, 2 * u, b_size, 50);
        paste_box(IminJ_Iy, panel, 26 * u, 8 * u, b_size, 50);

        /*
         *  Image and Displacement Vector Panel
         */

        // ROI and center
        cvRectangle(buf, cvPoint(b_xmin, b_ymin), cvPoint(b_xmax, b_ymax),
                    CV_RGB(0, 0, 0), 3);
        cvLine(buf, cvPoint(b_xcen, 0), cvPoint(b_xcen, buf->height - 1),
               CV_RGB(0, 0, 0), 3);
        cvLine(buf, cvPoint(0, b_ycen), cvPoint(buf->width - 1, b_ycen),
               CV_RGB(0, 0, 0), 3);

        //printf("G = [%f, %f; %f, %f]\n", G11, G12, G12, G22);
        printf("(eig1, eig2) = (%f, %f)\n", eig1, eig2);
        
        if (min_eig > EIG_THRESH) {
            // final solution (drawn in red)
            cvLine(buf,
                   cvPoint(b_xcen, b_ycen),
                   cvPoint(b_xcen + dx, b_ycen + dy),
                   CV_RGB(255, 0, 0), 3);

            // solutions per iteration (green cross)
            for (int i = 0; i < iter; i++) {
                cvLine(buf,
                       cvPoint(b_xcen + dx_traj[i] - 2, b_ycen + dy_traj[i]),
                       cvPoint(b_xcen + dx_traj[i] + 2, b_ycen + dy_traj[i]),
                       CV_RGB(0, 255, 0), 1);
                cvLine(buf,
                       cvPoint(b_xcen + dx_traj[i], b_ycen + dy_traj[i] - 2),
                       cvPoint(b_xcen + dx_traj[i], b_ycen + dy_traj[i] + 2),
                       CV_RGB(0, 255, 0), 1);
            }

            printf("(dx, dy) = (%f, %f)\n", dx, dy);
            printf("iter = %d\n", iter);
        }        

        cvShowImage("image", buf);
        cvShowImage("panel", panel);

        if (cvWaitKey(30) > 0) {
            break;
        }
    }

    return 0;
}
